"use strict";
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);
var create_exports = {};
__export(create_exports, {
  applyInitialState: () => applyInitialState,
  createState: () => createState
});
module.exports = __toCommonJS(create_exports);
var import_event = require("@domql/event");
var import_utils = require("@domql/utils");
var import_ignore = require("./ignore.js");
var import_methods = require("./methods.js");
var import_updateState = require("./updateState.js");
var import_inherit = require("./inherit.js");
const createState = async function(element, parent, options) {
  element.state = await applyInitialState(element, parent, options);
  return element.state;
};
const applyInitialState = async function(element, parent, options) {
  const objectizeState = await checkForTypes(element);
  if (objectizeState === false) return parent.state || {};
  else element.state = objectizeState;
  const whatInitReturns = await (0, import_event.triggerEventOn)("stateInit", element, options);
  if (whatInitReturns === false) return element.state;
  if ((0, import_inherit.checkIfInherits)(element)) {
    const inheritedState = (0, import_inherit.createInheritedState)(element, parent);
    element.state = (0, import_utils.isUndefined)(inheritedState) ? {} : inheritedState;
  }
  const dependentState = applyDependentState(element, element.state || parent.state || {});
  if (dependentState) element.state = dependentState;
  applyMethods(element);
  await (0, import_event.triggerEventOn)("stateCreated", element);
  return element.state;
};
const applyDependentState = (element, state) => {
  const { __ref, ref, __element } = state;
  const origState = (0, import_utils.exec)(__ref || ref || (__element == null ? void 0 : __element.state), element);
  if (!origState) return;
  const dependentState = (0, import_utils.deepClone)(origState, import_ignore.IGNORE_STATE_PARAMS);
  const newDepends = { [element.key]: dependentState };
  const __depends = (0, import_utils.isObject)(origState.__depends) ? { ...origState.__depends, ...newDepends } : newDepends;
  if (Array.isArray(origState)) {
    addProtoToArray(origState, { ...Object.getPrototypeOf(origState), __depends });
  } else {
    Object.setPrototypeOf(origState, { ...Object.getPrototypeOf(origState), __depends });
  }
  return dependentState;
};
const checkForTypes = async (element) => {
  const { state: orig, props, __ref: ref } = element;
  const state = (props == null ? void 0 : props.state) || orig;
  if ((0, import_utils.isFunction)(state)) {
    ref.__state = state;
    return await (0, import_utils.exec)(state, element);
  } else if ((0, import_utils.is)(state)("string", "number")) {
    ref.__state = state;
    return { value: state };
  } else if (state === true) {
    ref.__state = element.key;
    return {};
  } else if (state) {
    ref.__hasRootState = true;
    return state;
  } else {
    return false;
  }
};
const addProtoToArray = (state, proto) => {
  for (const key in proto) {
    Object.defineProperty(state, key, {
      value: proto[key],
      enumerable: false,
      // Set this to true if you want the method to appear in for...in loops
      configurable: true,
      // Set this to true if you want to allow redefining/removing the property later
      writable: true
      // Set this to true if you want to allow changing the function later
    });
  }
};
const applyMethods = (element) => {
  const state = element.state;
  const ref = element.__ref;
  const proto = {
    clean: import_methods.clean.bind(state),
    parse: import_methods.parse.bind(state),
    destroy: import_methods.destroy.bind(state),
    update: import_updateState.updateState.bind(state),
    rootUpdate: import_methods.rootUpdate.bind(state),
    parentUpdate: import_methods.parentUpdate.bind(state),
    create: createState.bind(state),
    add: import_methods.add.bind(state),
    toggle: import_methods.toggle.bind(state),
    remove: import_methods.remove.bind(state),
    apply: import_methods.apply.bind(state),
    applyReplace: import_methods.applyReplace.bind(state),
    applyFunction: import_methods.applyFunction.bind(state),
    set: import_methods.set.bind(state),
    quietUpdate: import_methods.quietUpdate.bind(state),
    replace: import_methods.replace.bind(state),
    quietReplace: import_methods.quietReplace.bind(state),
    reset: import_methods.reset.bind(state),
    parent: element.parent.state || state,
    setByPath: import_methods.setByPath.bind(state),
    setPathCollection: import_methods.setPathCollection.bind(state),
    removeByPath: import_methods.removeByPath.bind(state),
    removePathCollection: import_methods.removePathCollection.bind(state),
    getByPath: import_methods.getByPath.bind(state),
    keys: import_methods.keys.bind(state),
    values: import_methods.values.bind(state),
    __element: element,
    __children: {},
    root: ref.root ? ref.root.state : state
  };
  if ((0, import_utils.isArray)(state)) {
    addProtoToArray(state, proto);
  } else {
    Object.setPrototypeOf(state, proto);
  }
  if (state.parent && state.parent.__children) {
    state.parent.__children[element.key] = state;
  }
};
