"use strict";
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);
var inherit_exports = {};
__export(inherit_exports, {
  checkIfInherits: () => checkIfInherits,
  createInheritedState: () => createInheritedState,
  createNestedObjectByKeyPath: () => createNestedObjectByKeyPath,
  findInheritedState: () => findInheritedState,
  getChildStateInKey: () => getChildStateInKey,
  getParentStateInKey: () => getParentStateInKey,
  getRootStateInKey: () => getRootStateInKey,
  isState: () => isState
});
module.exports = __toCommonJS(inherit_exports);
var import_utils = require("@domql/utils");
var import_ignore = require("./ignore.js");
const getRootStateInKey = (stateKey, parentState) => {
  if (!stateKey.includes("~/")) return;
  const arr = stateKey.split("~/");
  if (arr.length > 1) return parentState.root;
};
const getParentStateInKey = (stateKey, parentState) => {
  if (!stateKey.includes("../")) return;
  const arr = stateKey.split("../");
  const arrLength = arr.length - 1;
  for (let i = 0; i < arrLength; i++) {
    if (!parentState.parent) return null;
    parentState = parentState.parent;
  }
  return parentState;
};
const getChildStateInKey = (stateKey, parentState, options = {}) => {
  const arr = stateKey.split("/");
  const arrLength = arr.length - 1;
  for (let i = 0; i < arrLength; i++) {
    const childKey = arr[i];
    const grandChildKey = arr[i + 1];
    if (childKey === "__proto__" || grandChildKey === "__proto__") return;
    let childInParent = parentState[childKey];
    if (!childInParent) childInParent = parentState[childKey] = {};
    if (!childInParent[grandChildKey]) childInParent[grandChildKey] = {};
    stateKey = grandChildKey;
    parentState = childInParent;
  }
  if (options.returnParent) return parentState;
  return parentState[stateKey];
};
const findInheritedState = (element, parent, options = {}) => {
  const ref = element.__ref;
  let stateKey = ref.__state;
  if (!checkIfInherits(element)) return;
  const rootState = getRootStateInKey(stateKey, parent.state);
  let parentState = parent.state;
  if (rootState) {
    parentState = rootState;
    stateKey = stateKey.replaceAll("~/", "");
  } else {
    const findGrandParentState = getParentStateInKey(stateKey, parent.state);
    if (findGrandParentState) {
      parentState = findGrandParentState;
      stateKey = stateKey.replaceAll("../", "");
    }
  }
  if (!parentState) return;
  return getChildStateInKey(stateKey, parentState, options);
};
const createInheritedState = (element, parent) => {
  const ref = element.__ref;
  const inheritedState = findInheritedState(element, parent);
  if ((0, import_utils.isUndefined)(inheritedState)) return element.state;
  if ((0, import_utils.is)(inheritedState)("object", "array")) {
    return (0, import_utils.deepClone)(inheritedState, { exclude: import_ignore.IGNORE_STATE_PARAMS });
  } else if ((0, import_utils.is)(inheritedState)("string", "number", "boolean")) {
    ref.__stateType = typeof inheritedState;
    return { value: inheritedState };
  }
  console.warn(ref.__state, "is not present. Replacing with", {});
};
const checkIfInherits = (element) => {
  const ref = element.__ref;
  const stateKey = ref.__state;
  if (stateKey && (0, import_utils.is)(stateKey)("number", "string", "boolean")) return true;
  return false;
};
const isState = function(state) {
  if (!(0, import_utils.isObjectLike)(state)) return false;
  return state.update && state.parse && state.clean && state.create && state.parent && state.destroy && state.rootUpdate && state.parentUpdate && state.keys && state.values && state.toggle && state.replace && state.quietUpdate && state.quietReplace && state.add && state.apply && state.applyReplace && state.setByPath && state.setPathCollection && state.removeByPath && state.removePathCollection && state.getByPath && state.applyFunction && state.__element && state.__children;
};
const createNestedObjectByKeyPath = (path, value) => {
  if (!path) {
    return value || {};
  }
  const keys = path.split("/");
  const obj = {};
  let ref = obj;
  keys.forEach((key, index) => {
    ref[key] = index === keys.length - 1 ? value || {} : {};
    ref = ref[key];
  });
  return obj;
};
