"use strict";
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);
var methods_exports = {};
__export(methods_exports, {
  add: () => add,
  apply: () => apply,
  applyFunction: () => applyFunction,
  applyReplace: () => applyReplace,
  clean: () => clean,
  destroy: () => destroy,
  getByPath: () => getByPath,
  keys: () => keys,
  parentUpdate: () => parentUpdate,
  parse: () => parse,
  quietReplace: () => quietReplace,
  quietUpdate: () => quietUpdate,
  remove: () => remove,
  removeByPath: () => removeByPath,
  removePathCollection: () => removePathCollection,
  replace: () => replace,
  reset: () => reset,
  rootUpdate: () => rootUpdate,
  set: () => set,
  setByPath: () => setByPath,
  setPathCollection: () => setPathCollection,
  toggle: () => toggle,
  values: () => values
});
module.exports = __toCommonJS(methods_exports);
var import_utils = require("@domql/utils");
var import_ignore = require("./ignore.js");
const parse = function() {
  const state = this;
  if ((0, import_utils.isObject)(state)) {
    const obj = {};
    for (const param in state) {
      if (!import_ignore.IGNORE_STATE_PARAMS.includes(param)) {
        obj[param] = state[param];
      }
    }
    return obj;
  } else if ((0, import_utils.isArray)(state)) {
    return state.filter((item) => !import_ignore.IGNORE_STATE_PARAMS.includes(item));
  }
};
const clean = function(options = {}) {
  const state = this;
  for (const param in state) {
    if (!import_ignore.IGNORE_STATE_PARAMS.includes(param) && Object.hasOwnProperty.call(state, param)) {
      delete state[param];
    }
  }
  if (!options.preventStateUpdate) {
    state.update(state, { replace: true, ...options });
  }
  return state;
};
const destroy = function(options = {}) {
  const state = this;
  const element = state.__element;
  const stateKey = element.__ref.__state;
  if ((0, import_utils.isString)(stateKey)) {
    element.parent.state.remove(stateKey, { isHoisted: true, ...options });
    return element.state;
  }
  delete element.state;
  element.state = state.parent;
  if (state.parent) {
    delete state.parent.__children[element.key];
  }
  if (state.__children) {
    for (const key in state.__children) {
      const child = state.__children[key];
      if (child.state) {
        if ((0, import_utils.isArray)(child.state)) {
          Object.defineProperty(child.state, "parent", {
            value: state.parent,
            enumerable: false,
            // Set this to true if you want the method to appear in for...in loops
            configurable: true,
            // Set this to true if you want to allow redefining/removing the property later
            writable: true
            // Set this to true if you want to allow changing the function later
          });
        } else {
          Object.setPrototypeOf(child, { parent: state.parent });
        }
      }
    }
  }
  element.state.update({}, { isHoisted: true, ...options });
  return element.state;
};
const parentUpdate = function(obj, options = {}) {
  const state = this;
  if (!state || !state.parent) return;
  return state.parent.update(obj, { isHoisted: true, ...options });
};
const rootUpdate = function(obj, options = {}) {
  const state = this;
  if (!state) return;
  const rootState = state.__element.__ref.root.state;
  return rootState.update(obj, { isHoisted: false, ...options });
};
const add = function(value, options = {}) {
  const state = this;
  if ((0, import_utils.isArray)(state)) {
    state.push(value);
    state.update(state.parse(), { overwrite: true, ...options });
  } else if ((0, import_utils.isObject)(state)) {
    const key = Object.keys(state).length;
    state.update({ [key]: value }, options);
  }
};
const toggle = function(key, options = {}) {
  const state = this;
  state.update({ [key]: !state[key] }, options);
};
const remove = function(key, options = {}) {
  const state = this;
  if ((0, import_utils.isArray)(state)) (0, import_utils.removeFromArray)(state, key);
  if ((0, import_utils.isObject)(state)) (0, import_utils.removeFromObject)(state, key);
  if (options.applyReset)
    return state.set(state.parse(), { replace: true, ...options });
  return state.update();
};
const set = function(val, options = {}) {
  const state = this;
  const value = (0, import_utils.deepClone)(val);
  return state.clean({ preventStateUpdate: true, ...options }).update(value, { replace: true, ...options });
};
const setByPath = function(path, val, options = {}) {
  const state = this;
  const value = (0, import_utils.deepClone)(val);
  if (!options.preventReplace) (0, import_utils.setInObjectByPath)(state, path, val);
  const update = (0, import_utils.createNestedObject)(path, value);
  if (options.preventStateUpdate) return update;
  return state.update(update, options);
};
const setPathCollection = async function(changes, options = {}) {
  const state = this;
  const update = changes.reduce((acc, change) => {
    if (change[0] === "update") {
      const result = setByPath.call(state, change[1], change[2], {
        ...options,
        preventStateUpdate: true
      });
      return (0, import_utils.overwriteDeep)(acc, result);
    } else if (change[0] === "delete") {
      removeByPath.call(state, change[1], options);
    }
    return acc;
  }, {});
  return await state.update(update, options);
};
const removeByPath = async function(path, options = {}) {
  const state = this;
  (0, import_utils.removeNestedKeyByPath)(state, path);
  if (options.preventUpdate) return path;
  return await state.update({}, options);
};
const removePathCollection = function(changes, options = {}) {
  const state = this;
  changes.forEach((item) => {
    removeByPath(item, { preventUpdate: true });
  });
  return state.update({}, options);
};
const getByPath = function(path, options = {}) {
  const state = this;
  return (0, import_utils.getInObjectByPath)(state, path);
};
const reset = function(options = {}) {
  const state = this;
  const value = (0, import_utils.deepClone)(state.parse());
  return state.set(value, { replace: true, ...options });
};
const apply = function(func, options = {}) {
  const state = this;
  if ((0, import_utils.isFunction)(func)) {
    const value = func(state);
    return state.update(value, { replace: true, ...options });
  }
};
const applyReplace = function(func, options = {}) {
  const state = this;
  if ((0, import_utils.isFunction)(func)) {
    const value = func(state);
    return state.replace(value, options);
  }
};
const applyFunction = function(func, options = {}) {
  const state = this;
  if ((0, import_utils.isFunction)(func)) {
    func(state);
    return state.update(state.parse(), { replace: true, ...options });
  }
};
const quietUpdate = function(obj, options = {}) {
  const state = this;
  return state.update(obj, { preventUpdate: true, ...options });
};
const replace = function(obj, options = {}) {
  const state = this;
  for (const param in obj) {
    state[param] = obj[param];
  }
  return state.update(obj, options);
};
const quietReplace = function(obj, options = {}) {
  const state = this;
  return state.replace(obj, { preventUpdate: true, ...options });
};
const keys = function(obj, options = {}) {
  const state = this;
  return Object.keys(state);
};
const values = function(obj, options = {}) {
  const state = this;
  return Object.values(state);
};
