"use strict";
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);
var updateState_exports = {};
__export(updateState_exports, {
  updateState: () => updateState
});
module.exports = __toCommonJS(updateState_exports);
var import_report = require("@domql/report");
var import_event = require("@domql/event");
var import_ignore = require("./ignore.js");
var import_utils = require("@domql/utils");
var import_inherit = require("./inherit.js");
const STATE_UPDATE_OPTIONS = {
  overwrite: true,
  preventHoistElementUpdate: false,
  updateByState: true,
  isHoisted: true,
  execStateFunction: true
};
const updateState = async function(obj, options = STATE_UPDATE_OPTIONS) {
  const state = this;
  const element = state.__element;
  if (options.onEach) options.onEach(element, state, element.context, options);
  if (!options.updateByState) (0, import_utils.merge)(options, STATE_UPDATE_OPTIONS);
  if (!state.__element) (0, import_report.report)("ElementOnStateIsNotDefined");
  if (options.preventInheritAtCurrentState === true) {
    options.preventInheritAtCurrentState = state;
  } else if (options.preventInheritAtCurrentState) return;
  if (!options.preventBeforeStateUpdateListener) {
    const beforeStateUpdateReturns = await (0, import_event.triggerEventOnUpdate)("beforeStateUpdate", obj, element, options);
    if (beforeStateUpdateReturns === false) return element;
  }
  applyOverwrite(state, obj, options);
  const updateIsHoisted = await hoistStateUpdate(state, obj, options);
  if (updateIsHoisted) return state;
  await updateDependentState(state, obj, options);
  await applyElementUpdate(state, obj, options);
  if (!options.preventStateUpdateListener) {
    await (0, import_event.triggerEventOnUpdate)("stateUpdate", obj, element, options);
  }
  return state;
};
const applyOverwrite = (state, obj, options) => {
  const { overwrite } = options;
  if (!overwrite) return;
  const shallow = overwrite === "shallow" || overwrite === "shallow-once";
  const merge2 = overwrite === "merge";
  if (merge2) {
    (0, import_utils.deepMerge)(state, obj, import_ignore.IGNORE_STATE_PARAMS);
    return;
  }
  const overwriteFunc = shallow ? import_utils.overwriteShallow : import_utils.overwriteDeep;
  if (options.overwrite === "shallow-once") options.overwrite = true;
  overwriteFunc(state, obj, import_ignore.IGNORE_STATE_PARAMS);
};
const hoistStateUpdate = async (state, obj, options) => {
  const element = state.__element;
  const { parent, __ref: ref } = element;
  const stateKey = ref == null ? void 0 : ref.__state;
  const stateType = ref == null ? void 0 : ref.__stateType;
  if (!stateKey) return;
  const asksForInherit = (0, import_inherit.checkIfInherits)(element);
  const inheritedState = (0, import_inherit.findInheritedState)(element, parent, { returnParent: true });
  const shouldPropagateState = asksForInherit && inheritedState && !options.stopStatePropagation;
  if (!shouldPropagateState) return;
  const isStringState = stateType === "string" || stateType === "number" || stateType === "boolean";
  const value = isStringState ? state.value : state.parse();
  const passedValue = isStringState ? state.value : obj;
  const findRootState = (0, import_inherit.getRootStateInKey)(stateKey, parent.state);
  const findGrandParentState = (0, import_inherit.getParentStateInKey)(stateKey, parent.state);
  const changesValue = (0, import_inherit.createNestedObjectByKeyPath)(stateKey, passedValue);
  const targetParent = findRootState || findGrandParentState || parent.state;
  if (options.replace) (0, import_utils.overwriteDeep)(targetParent, changesValue || value);
  await targetParent.update(changesValue, {
    execStateFunction: false,
    isHoisted: true,
    preventUpdate: options.preventHoistElementUpdate,
    overwrite: !options.replace,
    ...options
  });
  const hasNotUpdated = options.preventUpdate !== true || !options.preventHoistElementUpdate;
  if (!options.preventStateUpdateListener && hasNotUpdated) {
    await (0, import_event.triggerEventOnUpdate)("stateUpdate", obj, element, options);
  }
  return true;
};
const updateDependentState = async (state, obj, options) => {
  if (!state.__depends) return;
  for (const el in state.__depends) {
    const dependentState = state.__depends[el];
    await dependentState.clean().update(state.parse(), options);
  }
};
const applyElementUpdate = async (state, obj, options) => {
  const element = state.__element;
  if (options.preventUpdate !== true) {
    await element.update({}, {
      ...options,
      updateByState: true
    });
  } else if (options.preventUpdate === "recursive") {
    await element.update({}, {
      ...options,
      isHoisted: false,
      updateByState: true,
      preventUpdate: true
    });
  }
};
