var __defProp = Object.defineProperty;
var __defProps = Object.defineProperties;
var __getOwnPropDescs = Object.getOwnPropertyDescriptors;
var __getOwnPropSymbols = Object.getOwnPropertySymbols;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __propIsEnum = Object.prototype.propertyIsEnumerable;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp.call(b, prop))
      __defNormalProp(a, prop, b[prop]);
  if (__getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(b)) {
      if (__propIsEnum.call(b, prop))
        __defNormalProp(a, prop, b[prop]);
    }
  return a;
};
var __spreadProps = (a, b) => __defProps(a, __getOwnPropDescs(b));
import {
  isArray,
  deepClone,
  isFunction,
  isObject,
  isString,
  removeFromArray,
  removeFromObject,
  overwriteDeep,
  createNestedObject,
  getInObjectByPath,
  removeNestedKeyByPath,
  setInObjectByPath
} from "@domql/utils";
import { IGNORE_STATE_PARAMS } from "./ignore.js";
const parse = function() {
  const state = this;
  if (isObject(state)) {
    const obj = {};
    for (const param in state) {
      if (!IGNORE_STATE_PARAMS.includes(param)) {
        obj[param] = state[param];
      }
    }
    return obj;
  } else if (isArray(state)) {
    return state.filter((item) => !IGNORE_STATE_PARAMS.includes(item));
  }
};
const clean = function(options = {}) {
  const state = this;
  for (const param in state) {
    if (!IGNORE_STATE_PARAMS.includes(param) && Object.hasOwnProperty.call(state, param)) {
      delete state[param];
    }
  }
  if (!options.preventStateUpdate) {
    state.update(state, __spreadValues({ replace: true }, options));
  }
  return state;
};
const destroy = function(options = {}) {
  const state = this;
  const element = state.__element;
  const stateKey = element.__ref.__state;
  if (isString(stateKey)) {
    element.parent.state.remove(stateKey, __spreadValues({ isHoisted: true }, options));
    return element.state;
  }
  delete element.state;
  element.state = state.parent;
  if (state.parent) {
    delete state.parent.__children[element.key];
  }
  if (state.__children) {
    for (const key in state.__children) {
      const child = state.__children[key];
      if (child.state) {
        if (isArray(child.state)) {
          Object.defineProperty(child.state, "parent", {
            value: state.parent,
            enumerable: false,
            // Set this to true if you want the method to appear in for...in loops
            configurable: true,
            // Set this to true if you want to allow redefining/removing the property later
            writable: true
            // Set this to true if you want to allow changing the function later
          });
        } else {
          Object.setPrototypeOf(child, { parent: state.parent });
        }
      }
    }
  }
  element.state.update({}, __spreadValues({ isHoisted: true }, options));
  return element.state;
};
const parentUpdate = function(obj, options = {}) {
  const state = this;
  if (!state || !state.parent) return;
  return state.parent.update(obj, __spreadValues({ isHoisted: true }, options));
};
const rootUpdate = function(obj, options = {}) {
  const state = this;
  if (!state) return;
  const rootState = state.__element.__ref.root.state;
  return rootState.update(obj, __spreadValues({ isHoisted: false }, options));
};
const add = function(value, options = {}) {
  const state = this;
  if (isArray(state)) {
    state.push(value);
    state.update(state.parse(), __spreadValues({ overwrite: true }, options));
  } else if (isObject(state)) {
    const key = Object.keys(state).length;
    state.update({ [key]: value }, options);
  }
};
const toggle = function(key, options = {}) {
  const state = this;
  state.update({ [key]: !state[key] }, options);
};
const remove = function(key, options = {}) {
  const state = this;
  if (isArray(state)) removeFromArray(state, key);
  if (isObject(state)) removeFromObject(state, key);
  if (options.applyReset)
    return state.set(state.parse(), __spreadValues({ replace: true }, options));
  return state.update();
};
const set = function(val, options = {}) {
  const state = this;
  const value = deepClone(val);
  return state.clean(__spreadValues({ preventStateUpdate: true }, options)).update(value, __spreadValues({ replace: true }, options));
};
const setByPath = function(path, val, options = {}) {
  const state = this;
  const value = deepClone(val);
  if (!options.preventReplace) setInObjectByPath(state, path, val);
  const update = createNestedObject(path, value);
  if (options.preventStateUpdate) return update;
  return state.update(update, options);
};
const setPathCollection = async function(changes, options = {}) {
  const state = this;
  const update = changes.reduce((acc, change) => {
    if (change[0] === "update") {
      const result = setByPath.call(state, change[1], change[2], __spreadProps(__spreadValues({}, options), {
        preventStateUpdate: true
      }));
      return overwriteDeep(acc, result);
    } else if (change[0] === "delete") {
      removeByPath.call(state, change[1], options);
    }
    return acc;
  }, {});
  return await state.update(update, options);
};
const removeByPath = async function(path, options = {}) {
  const state = this;
  removeNestedKeyByPath(state, path);
  if (options.preventUpdate) return path;
  return await state.update({}, options);
};
const removePathCollection = function(changes, options = {}) {
  const state = this;
  changes.forEach((item) => {
    removeByPath(item, { preventUpdate: true });
  });
  return state.update({}, options);
};
const getByPath = function(path, options = {}) {
  const state = this;
  return getInObjectByPath(state, path);
};
const reset = function(options = {}) {
  const state = this;
  const value = deepClone(state.parse());
  return state.set(value, __spreadValues({ replace: true }, options));
};
const apply = function(func, options = {}) {
  const state = this;
  if (isFunction(func)) {
    const value = func(state);
    return state.update(value, __spreadValues({ replace: true }, options));
  }
};
const applyReplace = function(func, options = {}) {
  const state = this;
  if (isFunction(func)) {
    const value = func(state);
    return state.replace(value, options);
  }
};
const applyFunction = function(func, options = {}) {
  const state = this;
  if (isFunction(func)) {
    func(state);
    return state.update(state.parse(), __spreadValues({ replace: true }, options));
  }
};
const quietUpdate = function(obj, options = {}) {
  const state = this;
  return state.update(obj, __spreadValues({ preventUpdate: true }, options));
};
const replace = function(obj, options = {}) {
  const state = this;
  for (const param in obj) {
    state[param] = obj[param];
  }
  return state.update(obj, options);
};
const quietReplace = function(obj, options = {}) {
  const state = this;
  return state.replace(obj, __spreadValues({ preventUpdate: true }, options));
};
const keys = function(obj, options = {}) {
  const state = this;
  return Object.keys(state);
};
const values = function(obj, options = {}) {
  const state = this;
  return Object.values(state);
};
export {
  add,
  apply,
  applyFunction,
  applyReplace,
  clean,
  destroy,
  getByPath,
  keys,
  parentUpdate,
  parse,
  quietReplace,
  quietUpdate,
  remove,
  removeByPath,
  removePathCollection,
  replace,
  reset,
  rootUpdate,
  set,
  setByPath,
  setPathCollection,
  toggle,
  values
};
